# -*- coding: utf-8 -*-
"""Utilites for Cognite API SDK

This module provides helper methods and different utilities for the Cognite API Python SDK.

This module is protected and should not used by end-users.
"""
import json
import math
import re
import sys
import requests
import cognite._constants as _constants

def get_request(url, params=None, headers=None):
    '''Perform a GET request with a predetermined number of retries.'''
    for _ in range(_constants.RETRY_LIMIT + 1):
        res = requests.get(url, params=params, headers=headers)
        if res.status_code == 200:
            return res
    raise _APIError(res.json()['error'])

def post_request(url, body, headers=None):
    '''Perform a POST request with a predetermined number of retries.'''
    for _ in range(_constants.RETRY_LIMIT + 1):
        res = requests.post(url, data=json.dumps(body), headers=headers)
        if res.status_code == 200:
            return res
    raise _APIError(res.json()['error'])

def granularity_to_ms(time_string):
    '''Returns millisecond representation of granularity time string'''
    magnitude = int(''.join([c for c in time_string if c.isdigit()]))
    unit = ''.join([c for c in time_string if c.isalpha()])
    unit_in_ms = {'s': 1000, 'second': 1000, 'm': 60000, 'minute': 60000,
                  'h': 3600000, 'hour': 3600000, 'd': 86400000, 'day': 86400000}
    return magnitude * unit_in_ms[unit]

def _time_ago_to_ms(time_ago_string):
    '''Returns millisecond representation of time-ago string'''
    pattern = r'(\d+)([a-z])-ago'
    res = re.match(pattern, str(time_ago_string))
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        unit_in_ms = {'s': 1000, 'm': 60000, 'h': 3600000, 'd': 86400000, 'w': 604800000}
        return magnitude * unit_in_ms[unit]
    return None

class _APIError(Exception):
    pass

class ProgressIndicator():
    '''This class lets the system give the user and indication of how much data remains to be downloaded in the request'''
    def __init__(self, tag_ids, start, end):
        self.start, self.end = self._get_start_end(tag_ids, start, end)
        self.length = self.end - self.start
        self.progress = 0.0
        print("Downloading requested data...")
        self._print_progress()

    def update_progress(self, latest_timestamp):
        '''Update progress based on latest timestamp'''
        self.progress = 100 - (((self.end - latest_timestamp) / self.length) * 100)
        self._print_progress()

    def _print_progress(self):
        prog = int(math.ceil(self.progress) / 5)
        remainder = 20 - prog
        sys.stdout.write("\r{:5.1f}% |{}|".format(int(math.ceil(self.progress)), '|' * prog + ' ' * remainder))
        sys.stdout.flush()
        if int(math.ceil(self.progress)) == 100:
            print()

    def _get_start_end(self, tag_ids, start, end):
        from cognite.timeseries import get_latest, get_datapoints
        first_timestamp = float("inf")
        for tag in tag_ids:
            if isinstance(tag, str):
                tag_start = int(get_datapoints(tag, limit=1, start=start, end=end).to_json()['datapoints'][0]['timestamp'])
            else:
                tag_start = int(get_datapoints(tag['tagId'], limit=1).to_json()['datapoints'][0]['timestamp'])
            if tag_start < first_timestamp:
                first_timestamp = tag_start
        latest_timestamp = 0

        for tag in tag_ids:
            if isinstance(tag, str):
                tag_end = int(get_latest(tag).to_json()['timestamp'])
            else:
                tag_end = int(get_latest(tag['tagId']).to_json()['timestamp'])
            if tag_end > latest_timestamp:
                latest_timestamp = tag_end

        if start is None:
            self.start = first_timestamp
        elif _time_ago_to_ms(start): # start is specified as string of format '1d-ago'
            self.start = latest_timestamp - _time_ago_to_ms(start)
        else:
            self.start = start

        if end is None:
            self.end = latest_timestamp
        elif _time_ago_to_ms(end): # end is specified as string of format '1d-ago'
            self.end = latest_timestamp - _time_ago_to_ms(end)
        else:
            self.end = end

        return start, end



# -*- coding: utf-8 -*-
"""Timeseries Module

This module mirrors the Timeseries API. It allows you to fetch data from the api and output it in various formats.
"""
import io
import pandas as pd
import cognite.config as config
import cognite._constants as _constants
import cognite._utils as _utils

from cognite.data_objects import DatapointsObject, LatestDatapointObject

def get_datapoints(tag_id, aggregates=None, granularity=None, start=None, end=None, limit=_constants.LIMIT,
                   api_key=None, project=None):
    '''Returns a DatapointsObject containing a list of datapoints for the given query.

    Args:
        tag_id (str):            The tag_id to retrieve data for.

        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
        are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
        second/s', or a multiple of these indicated by a number as a prefix.

                                    Example: '12hour'.

        start (str):            Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.

                                    Example: '2d-ago' will get everything that is up to 2 days old. Can also send time in
                                    ms since epoch.

        end (str):              Get datapoints up to this time. Same format as for start.

        limit (int):            Return up to this number of datapoints.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        DatapointsObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(api_key, project)
    tag_id = tag_id.replace('/', '%2F')
    url = _constants.BASE_URL + '/projects/{}/timeseries/data/{}'.format(project, tag_id)
    params = {
        'aggregates': aggregates,
        'granularity': granularity,
        'limit': limit,
        'start': start,
        'end': end,
    }
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url, params=params, headers=headers)
    return DatapointsObject(res.json())

def get_latest(tag_id, api_key=None, project=None):
    '''Returns a LatestDatapointObject containing a list of datapoints for the given query.

    Args:
        tag_id (str):           The tag_id to retrieve data for.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        DatapointsObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = config.get_config_variables(api_key, project)
    tag_id = tag_id.replace('/', '%2F')
    url = _constants.BASE_URL + '/projects/{}/timeseries/latest/{}'.format(project, tag_id)
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url, headers=headers)
    return LatestDatapointObject(res.json())

def get_multi_tag_datapoints(tag_ids, aggregates=None, granularity=None, start=None, end=None, limit=_constants.LIMIT,
                             api_key=None, project=None):
    '''Returns a list of DatapointsObjects each of which contains a list of datapoints for the given tag_id.

    Args:
        tag_ids (list):         The list of tag_ids to retrieve data for.

        aggregates (list):      The list of aggregate functions you wish to apply to the data. Valid aggregate functions
        are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
        second/s', or a multiple of these indicated by a number as a prefix.

                                    Example: '12h'.

        start (str):            Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.

                                    Example: '2d-ago' will get everything that is up to 2 days old. Can also send time in
                                    ms since epoch.

        end (str):              Get datapoints up to this time. Same format as for start.

        limit (int):            Return up to this number of datapoints.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        list(DatapointsObjects): A list of data objects containing the requested data with several getter methods
        with different output formats.
    '''
    api_key, project = config.get_config_variables(api_key, project)
    url = _constants.BASE_URL + '/projects/{}/timeseries/dataquery'.format(project)
    body = {
        'items': [{'tagId': '{}'.format(tag_id)}
                  if isinstance(tag_id, str)
                  else {'tagId': '{}'.format(tag_id['tagId']), 'aggregates': tag_id['aggregates']} for tag_id in tag_ids],
        'aggregates': aggregates,
        'granularity': granularity,
        'start': start,
        'end': end,
        'limit': limit
    }
    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers)
    return [DatapointsObject({'data': {'items': [dp]}}) for dp in res.json()['data']['items']]

def get_datapoints_frame(tag_ids, aggregates, granularity, start=None, end=None, api_key=None, project=None):
    '''Returns a pandas dataframe of datapoints for the given tag_ids all on the same timestamps.

    Args:
        tag_ids (list):         The list of tag_ids to retrieve data for. Each tag_id can be either a string containing
        the tag_id or a dictionary containing the tag_id and a list of specific aggregate functions.

                                    Example option 1: ['<tag_id1>', '<tag_id2>'].

                                    Example option 2: [{'tagId': '<tag_id1>', 'aggregates': ['avg', 'max']},
                                                    {'tagId': '<tag_id2>', 'aggregates': ['step']}]

        aggregates (list):      The list of aggregate functions you wish to apply to the data for which you have not
        specified an aggregate function. Valid aggregate functions are: 'average/avg, max, min, count, sum, interpolation/int, stepinterpolation/step'.

        granularity (str):      The granularity of the aggregate values. Valid entries are : 'day/d, hour/h, minute/m,
        second/s', or a multiple of these indicated by a number as a prefix.

                                    Example: '12hour'.

        start (str):            Get datapoints after this time. Format is N[timeunit]-ago where timeunit is w,d,h,m,s.

                                    Example: '2d-ago' will get everything that is up to 2 days old. Can also send time in
                                    ms since epoch.

        end (str):              Get datapoints up to this time. Same format as for start.

        api_key (str):          Your api-key.

        project (str):          Project name.

    Returns:
        pandas.DataFrame: A pandas dataframe containing the datapoints for the given tag_ids. The datapoints for all the
        tag_ids will all be on the same timestamps.
    '''
    api_key, project = config.get_config_variables(api_key, project)
    url = _constants.BASE_URL + '/projects/{}/timeseries/dataframe'.format(project)
    body = {
        'items': [{'tagId': '{}'.format(tag_id)}
                  if isinstance(tag_id, str)
                  else {'tagId': '{}'.format(tag_id['tagId']), 'aggregates': tag_id['aggregates']} for tag_id in tag_ids],
        'aggregates': aggregates,
        'granularity': granularity,
        'start': start,
        'end': end,
        'limit': _constants.LIMIT
    }
    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'text/csv'
    }
    dataframes = []
    prog_ind = _utils.ProgressIndicator(tag_ids, start, end)
    while not dataframes or dataframes[-1].shape[0] == _constants.LIMIT:
        res = _utils.post_request(url=url, body=body, headers=headers)
        dataframes.append(
            pd.read_csv(io.StringIO(res.content.decode(res.encoding if res.encoding else res.apparent_encoding))))
        latest_timestamp = int(dataframes[-1].iloc[-1, 0])
        prog_ind.update_progress(latest_timestamp)
        body['start'] = latest_timestamp + _utils.granularity_to_ms(granularity)
    return pd.concat(dataframes).reset_index(drop=True)



# -*- coding: utf-8 -*-
"""Assets Module.

This module mirrors the Assets API.
"""
import cognite.config as config
import cognite._constants as _constants
import cognite._utils as _utils
from cognite.data_objects import AssetSearchObject

# Author: TK
def search_assets(description, api_key=None, project=None):
    '''Returns assets matching provided description.

    TODO:
        * TK: Enter description, args, and returns
    '''
    api_key, project = config.get_config_variables(api_key, project)
    url = _constants.BASE_URL + '/projects/{}/assets'.format(project)
    params = {
        'description': description,
    }
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url, params=params, headers=headers)

    return AssetSearchObject(res.json())


# Author: TK
def get_assets(asset_id=None, depth=None, api_key=None, project=None):
    '''Returns assets with provided assetId.

    TODO:
        * TK: Enter description, args, and returns
    '''
    api_key, project = config.get_config_variables(api_key, project)
    url = _constants.BASE_URL + '/projects/{}/assets/{}'.format(project, asset_id)
    params = {
        'depth': depth,
    }
    headers = {
        'api-key': api_key,
        'accept': 'application/json'
    }
    res = _utils.get_request(url, params=params, headers=headers)
    return AssetSearchObject(res.json())



# -*- coding: utf-8 -*-
'''Project Configuration Module.

This module allows you to set an api-key and a project for your python project.
'''

_CONFIG_API_KEY = ''
_CONFIG_PROJECT = ''

def configure_session(api_key='', project=''):
    '''Sets session variables.

    Args:
        api_key (str):  Api-key for current project.

        project (str):  Project name for current project.
    '''
    global _CONFIG_API_KEY, _CONFIG_PROJECT
    _CONFIG_API_KEY = api_key
    _CONFIG_PROJECT = project

def get_config_variables(api_key, project):
    '''Returns current project config variables unless other is specified.

    Args:
        api_key (str):  Other specified api-key.

        project (str):  Other specified project name.

    Returns:
        tuple: api-key and project name belonging to current project unless other is specified.
    '''
    if api_key is None:
        api_key = _CONFIG_API_KEY
    if project is None:
        project = _CONFIG_PROJECT
    return api_key, project



# -*- coding: utf-8 -*-
'''Data Objects

This module contains data objects used to represent the data returned from the API. These objects have at least
the following output formats:

    * to_pandas():    Returns pandas dataframe
    * to_ndarray():   Numpy array
    * to_json():      Json format
'''
from abc import ABC, abstractmethod
import pandas as pd

# Author: Erlend Vollset
class CogniteDataObject(ABC):
    '''Abstract Cognite Data Object

    This abstract class provides a skeleton for all data objects in this module. All data objects should inherit
    this class.
    '''
    def __init__(self, internal_representation):
        self.internal_representation = internal_representation
    @abstractmethod
    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        pass
    @abstractmethod
    def to_json(self):
        '''Returns data as a json object'''
        pass

    def to_ndarray(self):
        '''Returns data as a numpy array'''
        return self.to_pandas().values

# Author: Erlend Vollset
class TagMatchingObject(CogniteDataObject):
    '''Tag Matching Data Object.

    In addition to the standard output formats this data object also has a to_list() method which returns a list of
    names of the tag matches.
    '''
    def __init__(self, internal_representation):
        CogniteDataObject.__init__(self, internal_representation)

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        matches = []
        for tag in self.internal_representation['data']['items']:
            for match in tag['matches']:
                matches.append({
                    'tag': tag['tagId'],
                    'match': match['tagId'],
                    'score': match['score'],
                    'platform': match['platform']
                })
        if matches:
            return pd.DataFrame(matches)[['tag', 'match', 'platform', 'score']]
        return pd.DataFrame()

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items']

    def to_list(self, first_matches_only=True):
        '''Returns a list representation of the matches.

        Args:
            first_matches_only (bool):      Boolean determining whether or not to return only the top match for each
                                            tag.

        Returns:
            list: list of matched tags.
        '''
        if self.to_pandas().empty:
            return []
        if first_matches_only:
            return self.to_pandas().sort_values(['score', 'match']).groupby(['tag']).first()['match'].tolist()
        return self.to_pandas().sort_values(['score', 'match'])['match'].tolist()


# Author: Erlend Vollset
class DatapointsObject(CogniteDataObject):
    '''Datapoints Object.'''
    def __init__(self, internal_representation):
        CogniteDataObject.__init__(self, internal_representation)

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items'][0]

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        return pd.DataFrame(self.internal_representation['data']['items'][0]['datapoints'])


# Author: Erlend Vollset
class LatestDatapointObject(CogniteDataObject):
    '''Latest Datapoint Object.'''
    def __init__(self, internal_representation):
        CogniteDataObject.__init__(self, internal_representation)

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items'][0]

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        return pd.DataFrame([self.internal_representation['data']['items'][0]])

    def to_ndarray(self):
        '''Returns data as a numpy array'''
        return self.to_pandas().values[0]


# Author: Erlend Vollset
class SimilaritySearchObject(CogniteDataObject):
    '''Similarity Search Data Object.'''
    def __init__(self, internal_representation):
        CogniteDataObject.__init__(self, internal_representation)

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items']

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        if len(self.to_json()) > 1:
            return pd.DataFrame(self.internal_representation['data']['items'])
        return pd.DataFrame()


# Author: T.K
class AssetSearchObject(CogniteDataObject):
    '''Assets Search Data Object

    TODO:
        * For now just a copy of SimilaritySearchObject. Implement correct formatting for this data.
    '''
    def __init__(self, internal_representation):
        CogniteDataObject.__init__(self, internal_representation)

    def to_json(self):
        '''Returns data as a json object'''
        return self.internal_representation['data']['items']

    def to_pandas(self):
        '''Returns data as a pandas dataframe'''
        if len(self.to_json()) > 1:
            return pd.DataFrame(self.internal_representation['data']['items'])
        return pd.DataFrame()



from setuptools import setup

version = '0.3.8'

setup(name='cognite-sdk',
      version=version,
      description='Cognite API SDK for Python',
      url='https://cognitedata.github.io/cognite-sdk-python/',
      download_url='https://github.com/cognitedata/cognite-sdk-python/cognite/archive/{}.tar.gz'.format(version),
      author='Erlend Vollset',
      author_email='erlend.vollset@cognite.com',
      packages=['cognite'],
      install_requires=[
          'requests',
          'pandas'
      ],
      zip_safe=False,
      include_package_data=True)


'''Cognite API SDK

This package has been created to ensure excellent excellent user experience for data scientists using the Cognite
Data Platform (CDP).
'''

__all__ = ['assets', 'config', 'similarity_search', 'tagmatching', 'timeseries']



# -*- coding: utf-8 -*-
"""Similarity Search Module.

This module mirrors the Similarity Search API.
"""
import cognite._constants as _constants
import cognite._utils as _utils
from cognite.config import get_config_variables
from cognite.data_objects import SimilaritySearchObject

def search(input_tags, query_tags, input_interval, query_interval, modes, limit=10, api_key=None, project=None):
    '''Returns patterns within query interval which are similar to input pattern.

    This method allows you to enter an input interval (the desired interval of a timeseries to use as input) and a
    query interval (the desired interval to search for similarities in) and will return the list of intervals which
    are most similar to your input.

    Args:
        input_tags (list):      List of tags corresponding to time-series patterns to search.

        query_tags (list):      List of tags of time-series that we want to search similar patterns on.

        input_interval (tuple): Start and end of input interval.

                                    Example: {'start': 0, 'end: 100}.

        query_interval (tuple): Start and end of input interval.

                                    Example: {'start': 0, 'end: 100}.

        modes (list):           List of modes of searching. Modes could be 'value' or 'pattern'.

        limit (int):            Number of matches to retrieve.

    Returns:
        SimilaritySearchObject: A data object containing the requested data with several getter methods with different
        output formats.
    '''
    api_key, project = get_config_variables(api_key, project)
    url = _constants.BASE_URL + '/projects/{}/similaritysearch'.format(project)
    body = {
        'inputTags': input_tags,
        'queryTags': query_tags,
        'inputInterval': {'start': input_interval[0],
                          'end': input_interval[1]} if isinstance(input_interval, tuple) else input_interval,
        'queryInterval': {'start': query_interval[0],
                          'end': query_interval[1]} if isinstance(query_interval, tuple) else query_interval,
        'modes': modes,
        'limit': limit
    }
    headers = {
        'api-key': api_key,
        'content-type': 'application/json',
        'accept': 'application/json'
    }
    res = _utils.post_request(url=url, body=body, headers=headers)
    return SimilaritySearchObject(res.json())
